#include "Shared_string.h"
#include <catch.h>

TEST_CASE("empty")
{
    Shared_string s;

    CHECK( s.empty() );
    CHECK( s.size() == 0 );
    CHECK( s == "" );
}

TEST_CASE("simple construct-delete")
{
    Shared_string s("hello");

    CHECK_FALSE( s.empty() );
    CHECK( s.size() == 5 );
    CHECK( s == "hello" );
}

TEST_CASE("copy constructor and copy assignment")
{
    char const s0[] = "hello";
    Shared_string s1(s0);
    Shared_string s2(s1);
    Shared_string s3(s0);

    CHECK( s0 == s1 );
    CHECK( s0 == s2 );
    CHECK( s0 == s3 );
    CHECK( s1 == s0 );
    CHECK( s1 == s1 );
    CHECK( s1 == s2 );
    CHECK( s1 == s3 );
    CHECK( s2 == s0 );
    CHECK( s2 == s1 );
    CHECK( s2 == s2 );
    CHECK( s2 == s3 );
    CHECK( s3 == s0 );
    CHECK( s3 == s1 );
    CHECK( s3 == s2 );
    CHECK( s3 == s3 );

    CHECK( s1.begin() == s2.begin() );
    CHECK( s1.begin() != s3.begin() );
    CHECK( s2.begin() != s3.begin() );

    s2 = s3;

    CHECK( s1.begin() != s2.begin() );
    CHECK( s1.begin() != s3.begin() );
    CHECK( s2.begin() == s3.begin() );

    s1 = std::move(s2);

    CHECK( s2.empty() );
    CHECK_FALSE( s1.empty() );
    CHECK( s1.begin() == s3.begin() );

    s1 = Owned_string(s0);

    CHECK( s1.begin() != s2.begin() );
    CHECK( s1.begin() != s3.begin() );
    CHECK( s2.begin() != s3.begin() );

    s2 = s3 = s1;

    CHECK( s1.begin() == s2.begin() );
    CHECK( s1.begin() == s3.begin() );
    CHECK( s2.begin() == s3.begin() );
}

TEST_CASE("refcount")
{
    Shared_string s1;
    CHECK( s1.ref_count() == 0 );
    s1 = "hello";
    CHECK( s1.ref_count() == 1 );
    Shared_string s2 = s1;
    CHECK( s1.ref_count() == 2 );
    Shared_string s3 = s1;
    CHECK( s1.ref_count() == 3 );

    s3 = "";
    CHECK( s1.ref_count() == 2 );
    CHECK( s2.ref_count() == 2 );
    CHECK( s3.ref_count() == 0 );
    s1 = Shared_string();
    CHECK( s1.ref_count() == 0 );
    CHECK( s2.ref_count() == 1 );
    CHECK( s3.ref_count() == 0 );
}

TEST_CASE("substring")
{
    Shared_string s1("hello, world!");
    CHECK( s1.ref_count() == 1 );

    Shared_string s2 = s1.substring(7);
    CHECK( s1.ref_count() == 2 );
    CHECK( s2 == "world!" );

    Shared_string s3 = s2.substring(1, 3);
    CHECK( s1.ref_count() == 3 );
    CHECK( s3 == "orl" );

    s2 = "moo";
    CHECK( s1.ref_count() == 2 );
    CHECK( s3 == "orl" );
}

TEST_CASE("stealing into owned_string")
{
    Shared_string s1("hello");
    Shared_string s2(s1);

    const char* old_s1 = s1.begin();

    Owned_string s3 = std::move(s1);
    s1 = "";  // move may have no effect

    CHECK( s3 == "hello" );
    CHECK( old_s1 == s2.begin() );
    CHECK( old_s1 != s3.begin() );

    Owned_string s4 = std::move(s2);
    CHECK( old_s1 != s1.begin() );
    CHECK( old_s1 == s4.begin() );
    CHECK( s2.empty() );

    s1 = "hello, world";
    s1 = s1.substring(3, 5);
    Owned_string s5 = std::move(s1);  // no effect because s1 is a substring
    CHECK( s1 == "lo, w" );
    CHECK( s1 == s5 );
}
